$include(maxq3120.inc)
$include(codec_ram.inc)
;
; A-law and u-law codec based on the MAXQ3120 microcontroller
;
; Copyright (c) 2005, Dallas Semiconductor/Maxim
;
;================================================================================
;
; NOTICE
;
; This code is provided AS-IS, and no warranty of any kind is expressed or
; implied. This code has been verified to compile under the MAXQASM assembler,
; but is otherwise UNTESTED. Its intended purpose is to provide design ideas
; to the interested engineer. If you choose to develop this code into a product,
; you do so ENTIRELY AT YOUR OWN RISK, and YOU ARE RESPONSIBLE FOR VERIFYING THE
; FUNCTIONALITY OF THE FINAL PRODUCT.
;
;================================================================================
;
; Here's the idea:
;
; Inside timer interrupt:
;
;    1. Wait for frame pulse
;    2. Emit bit from A7.
;    3. Receive bit into A6.
;    4. Repeat steps 2-3 8 times.
;    5. Repeat steps 2-4 twice.
;    5. On conclusion, set done bit.
;
; Outside, in the main loop:
;
;    1. While waiting for done bit, scan ADC converters, adding samples
;       to the circular buffer as required.
;    2. When the done bit comes along, perform filtering, producing a 16-
;       bit linear sample.
;    3. Convert the linear sample to PCM by performing a binary search of
;       the PCM tables.
;    4. Move the new sample to A7.
;    5. Convert the A6 value to linear by direct table lookup (with sign bit
;       accountability, of course.
;    6. Clock the 16 bits to the DAC.
;    7. Lather, rinse, repeat.
;
;
; ACCUMULATOR USAGE:
;
; A0: Real arithmetic accumulator for non-interrupt operations.
; A1: General scratch register
; A2: Search value for CODEC_SEARCH routine
; A3: Not used in this design
; A4: Filtered transmit value, channel 0
; A5: Filtered transmit value, channel 1
; A6: Inbound companded value, two bytes.
; A7: Outbound companded value, two bytes.
; A8-A15: Not used in this design.
;
; DATA POINTER USAGE:
; DP0: Points to NEXT space in 32-word transmit buffer. Toggle bit 4 to select codec A or B.
; DP1: Points to codec table.
; BP[OFFS]: Not used in this design.
;
; System control bits usage:
;
; The MAXQ architecture uses a set of bits to map certain memory segments to defined places
; in the memory map. The MAXQ3120 has a smaller flash size, and so doesn't need these bits.
; Nevertheless, they're present in the 3120, and we can use them as we please. In this
; design, two of the bits are used: SC.4 defines the encoding type (0= MU-LAW, 1= A-LAW) and
; SC.3 serves as the DONE bit (set when a frame has been processed by the interrupt routine.)
;
;
;---------------------------------------------------------------------------------------------
; START OF MAIN
;
	org	0
main_init:
	move	ADCN, #30h			;Flush FIFOs
	move	ADCN, #300h			;Power down modulators
	move	ADCN, #03C0h		;Set gain factors, dual mod interrupt, 1X
	move	ADCN, #00C0h		;Turn 'er up!
	move	IV, #timer_isr		;Point to interrupt service routine
	move	DPC, #1Ch			;Set all data pointers for WORD operation
	move	PD3, #70h			;Set DAC interface bits to output mode
	move	PD2, #1				;Set PCM data out bit to output mode
;
; Before turning up the interrupts, check the pin for which codec table to load
; into RAM, and load the table. TO CHANGE THE TYPE OF CODEC, YOU MUST CHANGE THE
; STATE OF THE PIN AND RESET THE PROCESSOR!
;
	move	C, PI2.4			;Port 2.4 selects encoding: 0=mu, 1=a-law
	move	SC.4, #0			;Assume we're using A-law
	move	DP[1], #A_LAW
	sjump	C, use_alaw_table	;If external bit is set, then it *is* A-law
	move	DP[1], #MU_LAW		;Otherwise, load mu-law table address
	move	SC.4, #1
use_alaw_table:
	move	DP[0], #800dh		;Point to table of routines
	move	ACC, @DP[0]			;Get table pointer
	add		#13					;Add offset to table move routine
	move	DP[0], ACC			;Move offset pointer to DP[0]
	move	ACC, @DP[0]			;Get actual address of routine
	move	DP[0], #codec_table	;Point to table area in RAM
	move	LC[0], #128			;Move 128 entries
	call	ACC					;Call the table move routine
;
; Now the table has been loaded...we can turn up the interrupts.
;
	move	IM1, #1				;IM1 enables module 1 interrupts (only timer 0 enabled)
	move	IGE, #1				;IGE globally enables interrupts.
	move	T0L, #900 & 0FFh
	move	T0H, #900 << 8		;Load 900 into timer
	move	ET0, #1				;Enable the timer interrupt
;
; Timer is running. Scan DONE bit and ADC sample ready bits
;
; First, check DONE bit
;
main_loop:
	move	C, SC.3				;SC.3 is the done bit.
	sjump	NC, main_loop_end	;If it's not set, skip end-of-frame processing.
	move	SC.3, #0			;Otherwise, clear the done bit and proceed with processing.
;
; If done bit is not set, skip this section. Otherwise...
; Begin with receive tasks. Find the appropriate 16-bit value in the RAM table, and
; send it to the DAC channel.
;
; Here's how this works: the sixteen-bit value received from the PCM bus is in A[7]. It
; consists of two eight-bit samples. We'll use GR to seperate them. Each sample must be
; converted to a twelve-bit linear form by looking up in the codec table, and then output
; to the external DAC.
;
	move	GR, A[7]			;A7 has the recevied samples, use GR to split them.
	move	ACC, GRL			;GRL has channel 0.
	scall	law_adjust_in		;LAW_ADJUST_IN has the logic to undo the law weirdness
	scall	lookup_codec_value	;This converts the un-weird value to a 12-bit linear value
	scall	DAC_send			;Send 12-bit signed value to the DAC, channel 0
	move	ACC, GRH			;Do the same thing for channel 1.
	scall	law_adjust_in		;Adjust channel 1 value
	scall	lookup_codec_value	;Lookup channel 1 linear value
	add		#1000h				;This sets the channel 1 bit in the DAC
	scall	DAC_send			;Finally, output the channel 1 receive value.

;
; Receive task is complete. Transmit is a little trickier. First, perform the filter function
; on the samples received to date.
;
; Here's what's going on: the two ADC channels keep 32 samples each in a pair of circular
; buffers. To match the frame rate to the sample rate (8 ksps vs. 20.667 ksps) we use an FIR
; filter algorithm. It's a 31 order, Hamming-window filter at 3.5 kHz.
;
	call	check_adc_channels	;Before filtering, check if ADC needs attention
	scall	fir_filter			;This filters the data at DP0 and leaves the result in MC
	move	A[4], MC1			;A4 has the transmit value for channel 0.
	move	A[1], DP[0]			;We're about to do violence to DP0, so save it.
	move	ACC, DP[0]			;Get the pointer for adjustment
	or		#20h				;Point up to the next table
	move	DP[0], ACC			;...and save the result back to the pointer register
	call	check_adc_channels	;Before filtering, check if ADC needs attention
	scall	fir_filter			;Now, filter channel 1 values
	move	A[5], MC1			;A5 contains the 16-bit linear tx value for ch1
	move	DP[0], A[1]			;Finally, restore the old pointer.
;
; FIR filtering is complete, and the results are in A4 and A5.
; Now find the closest compressed sample value to the filter value.
;
; For each channel, we have to determine if the value is negative. If it is, we have
; to invert it, *then* call CODEC_SEARCH, and finally apply a sign bit.
;
	move	ACC, A[4]			;Get filtered value for channel 0.
	sjump	S, sample_0_negative	;If negative, just jump
	move	A[2], A[4]			;Otherwise, if positive, move search value to A2...
	scall	codec_search		;...and call the search routine.
	move	ACC, DP[1]			;DP1 contains a pointer to the codec value, but...
	sub		#codec_table		;...it's the index we're after. Subtract the base.
	scall	law_adjust_out		;Apply A or MU law weirdness
	move	GRL, ACC			;This is the unsigned value (but we know it's positive.)
	sjump	test_sample_1		;Now, go test sample 1.
sample_0_negative:
	neg							;Sample 0 was negative...make it positive before processing.
	move	A[2], ACC			;Move sign-corrected value to A2 for processing
	scall	codec_search
	move	ACC, DP[1]			;DP1 contains a pointer to the codec value, but...
	sub		#codec_table		;...it's the index we're after. Subtract the base.
	or		#80h				;Now, set the sign bit (it's not 2's complement, but...
	scall	law_adjust_out		;...signed magnitude. Apply A or MU law weirdness.
	move	GRL, ACC			;
test_sample_1:
	move	ACC, A[5]			;Get filtered value for channel 1.
	sjump	S, sample_1_negative	;If negative, just jump
	move	A[2], A[5]			;Otherwise, if positive, move search value to A2
	scall	codec_search
	move	ACC, DP[1]			;DP1 contains a pointer to the codec value, but...
	sub		#codec_table		;...it's the index we're after. Subtract the base.
	scall	law_adjust_out		;Apply A or MU law weirdness
	move	GRH, ACC			;This is the unsigned value (but we know it's positive.)
	sjump	main_loop_end
sample_1_negative:
	neg							;Sample 0 was negative...make it positive before processing.
	move	A[2], ACC			;Move sign-corrected value to A2 for processing
	scall	codec_search
	move	ACC, DP[1]			;DP1 contains a pointer to the codec value, but...
	sub		#codec_table		;...it's the index we're after. Subtract the base.
	or		#80h				;Now, set the sign bit (it's not 2's complement, but...
	scall	law_adjust_out		;Apply A or MU law weirdness
	move	GRH, ACC
main_loop_end:
	move	A[6], GR			;GR contains the adjusted output sample set
	call	check_adc_channels	;As a last check, look at the ADC again.
	sjump	main_loop

;---------------------------------------------------------------------------------------------
;
;
; LAW_ADJUST
;
; Mu law and A law have different bizarro rules concerning how they send fairly simple
; values on the wire. This must be accounted for before applying the (rather simple)
; conversions between linear PCM and compressed PCM.
;
; There are two routines here: LAW_ADJUST_IN converts received compressed PCM samples
; anc converts them to signed magnitude compressed PCM. It puts the sign bit into C
; for easier processing by the caller. LAW_ADJUST_OUT takes signed magnitude samples
; from the codec search routine and converts them to lawful samples.
;
law_adjust_in:
	move	C, SC.4				;SC.4 contains the law bit: 1=MU, 0=A
	jump	C, adjust_mu_law_in
;
; We need to adjust A-law.
;
	xor		#0D5h				;A-law inverts every other bit and...
	sjump	adjust_A_law_in		;...the sign bit set indicates positive.
adjust_mu_law_in:
	cpl							;Mu law complements everything.
adjust_a_law_in:
	move	C, ACC.7			;Now, get the sign bit...
	and		#7fh				;...and make the return value UNSIGNED.
	ret

law_adjust_out:
	move	C, SC.4				;SC.4 contains the law bit: 1=MU, 0=A
	jump	C, adjust_mu_law_out
;
; We need to adjust A-law.
;
	xor		#0D5h				;A-law inverts every other bit and...
	sjump	adjust_A_law_out	;...the sign bit indicates positive.
adjust_mu_law_out:
	xor		#0FFh				;Mu law complements everything.
adjust_a_law_out:
	ret

;---------------------------------------------------------------------------------------------
;
;
; LOOKUP_CODEC_VALUE
;
; Input: ACC contains the adjusted 8-bit compressed value; that is,
;        it's really a 7-bit compressed magnitude. A1.0 contains
;        the sign.
;
; Output: ACC contains the 12-bit magnitude centered about 2048.
;
lookup_codec_value:
	add		#codec_table		;Just directly add the offset to the table address
	move	DP[1], ACC			;Move the offseted value to the pointer
	move	ACC, A[1]			;Get sign into the accumulator
	rrc							;Move sign bit into carry
	move	ACC, @DP[1]			;Now, get the 16-bit magnitude
	sjump	NC, do_not_negate	;If positive, don't negate the value
	neg							;Otherwise, make value negative
do_not_negate:
	sra4						;Save only 12 MSBs
	add		#2048				;Add zero offset
	and		#0FFFh				;Clear potential garbage
	ret

;---------------------------------------------------------------------------------------------
;
;
; CODEC_SEARCH:
;
; Performs a binary search through the 128 codec values that have been loaded into RAM.
;
; Input: MC contains a filtered output value. The job is to find the
;        nearest compressed value to send to the PCM bus.
;
; Output: DP[1] contains a pointer to the nearest value. The index is the
;         output value.
;
codec_search:
	move	A[1], #32			;Load *next* midpoint adjust
	move	DP[1], #codec_table + 64	;Point to midpoint
codec_search_loop:
	move	ACC, @DP[1]			;Get a codec value
	sub		A[2]				;Subtract input value; C set if input val>table val
	sjump	C, codec_search_too_small	;If C, need to search *higher* in the table.
	sjump	Z, codec_search_found	;If exactly equal, you found the match!
	move	ACC, DP[1]			;Need to search lower in the table.
	sub		A[1]				;Subtract binary search value...
	move	DP[1], ACC			;...and store pointer back.
	sjump	codec_search_continue
codec_search_too_small:
	move	ACC, DP[1]			;Need to search higher in the table
	add		A[1]				;Add binary search value
	move	DP[1], ACC			;...and store pointer back.
codec_search_continue:
	move	ACC, A[1]			;Now we need to adjust binary search value
	sra							;Divide by 2
	sjump	Z, codec_search_found	;If the division result=0, I'm done.
	move	A[1], ACC			;Otherwise, save bin search value and loop back for more
	sjump	codec_search_loop
codec_search_found:
	ret
;---------------------------------------------------------------------------------------------
;
;
; DAC_SEND
;
; Send the 16-bit value in ACC to the DAC.
;
; LC0 is used; ACC is destroyed.
;
DAC_send:
	move	PO3.5, #0			;Set CE to the DAC
	move	LC[0], #16			;We'll do 16 bits
DAC_send_loop:
	rlc							;Get a sample bit into C
	sjump	C, DAC_send_1		;Jump if I'm sending a "1"
	move	PO3.6, #0			;Set data to 0
	sjump	C, DAC_send_0
DAC_send_1:
	move	PO3.6, #1			;Set data to 1
DAC_send_0:
	move	PO3.4, #1			;Raise clock
	nop							;Wait 1 cycle
	move	PO3.4, #0			;Drop clock
	djnz	LC[0], DAC_send_loop	;Loop if there are more to send
	move	PO3.5, #1			;When done, deassert CE
	ret
;---------------------------------------------------------------------------------------------
;
;
; FIR_FILTER
; Filters the values in the buffer, leaving the result in MC1.
; DP0 points to the *next* "empty" location in the buffer. That is, it points to the
; oldest sample. I don't want the oldest samples -- I want the most recent samples.
; With an order 31 filter and 32 samples, I just increment the pointer by 1 to begin.
;
; The only hard part is that I have to check for overrun on the buffer pointer. The
; buffer has overrun if bits 0-4 are all zero after the pointer is incremented.
; Instead of testing the pointer (which can't be *directly* tested, and must be loaded
; into an accumulator for testing) I'm loading the complement into a loop counter. Now,
; I can use the DJNZ instruction to skip reloading the pointer most of the time.
;
; For example, if the buffer pointer contains 1Eh (two more buffer places to go before
; overrun) I take the complement and mask only the lowest 5 bits, leaving 2 -- just
; what I'm after! The only place I have to take care is if the result is zero -- if
; it is, then LC0 will get zero -- and will never go to zero (no pointer adjustments
; will take place!) But that's okay -- that means the pointer was at zero to begin
; with, and that's the ONE case where no adjustment IS necessary!
;
; Yes, this could be implemented as a loop quite easily...but it is unrolled here
; for speed. Note that only three instructions are typically used for each point
; of the filter, making this a very efficient implementation.
;
fir_filter:
	move	MCNT, #22h			;Clear data registers, signed operation.
	move	ACC, DP[0]			;Get buffer pointer
	neg							;Take the complement...
	and		#1Fh				;...and mask.
	move	LC[0], ACC			;Move result to the loop counter.
	move	ACC, @DP[0]++		;Now increment the pointer past the first item
	djnz	LC[0], no_rollover_init	;If LC0 != 0, skip next code section.
	move	ACC, DP[0]			;Get pointer
	and		#0E0h				;Mask pointer back to zero (range=0-1F or 20-3F)
	move	DP[0], acc			;Reload pointer
no_rollover_init:				;...and so forth.
	move	MA, @DP[0]++
	move	MB, #coefficient_0
	djnz	LC[0], no_rollover_00
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_00:
	move	MA, @DP[0]++
	move	MB, #coefficient_1
	djnz	LC[0], no_rollover_01
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_01:
	move	MA, @DP[0]++
	move	MB, #coefficient_2
	djnz	LC[0], no_rollover_02
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_02:
	move	MA, @DP[0]++
	move	MB, #coefficient_3
	djnz	LC[0], no_rollover_03
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_03:
	move	MA, @DP[0]++
	move	MB, #coefficient_4
	djnz	LC[0], no_rollover_04
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_04:
	move	MA, @DP[0]++
	move	MB, #coefficient_5
	djnz	LC[0], no_rollover_05
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_05:
	move	MA, @DP[0]++
	move	MB, #coefficient_6
	djnz	LC[0], no_rollover_06
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_06:
	move	MA, @DP[0]++
	move	MB, #coefficient_7
	djnz	LC[0], no_rollover_07
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_07:
	move	MA, @DP[0]++
	move	MB, #coefficient_8
	djnz	LC[0], no_rollover_08
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_08:
	move	MA, @DP[0]++
	move	MB, #coefficient_9
	djnz	LC[0], no_rollover_09
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_09:
	move	MA, @DP[0]++
	move	MB, #coefficient_10
	djnz	LC[0], no_rollover_10
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_10:
	move	MA, @DP[0]++
	move	MB, #coefficient_11
	djnz	LC[0], no_rollover_11
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_11:
	move	MA, @DP[0]++
	move	MB, #coefficient_12
	djnz	LC[0], no_rollover_12
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_12:
	move	MA, @DP[0]++
	move	MB, #coefficient_13
	djnz	LC[0], no_rollover_13
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_13:
	move	MA, @DP[0]++
	move	MB, #coefficient_14
	djnz	LC[0], no_rollover_14
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_14:
	move	MA, @DP[0]++
	move	MB, #coefficient_15
	djnz	LC[0], no_rollover_15
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_15:
	move	MA, @DP[0]++
	move	MB, #coefficient_14
	djnz	LC[0], no_rollover_16
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_16:
	move	MA, @DP[0]++
	move	MB, #coefficient_13
	djnz	LC[0], no_rollover_17
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_17:
	move	MA, @DP[0]++
	move	MB, #coefficient_12
	djnz	LC[0], no_rollover_18
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_18:
	move	MA, @DP[0]++
	move	MB, #coefficient_11
	djnz	LC[0], no_rollover_19
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_19:
	move	MA, @DP[0]++
	move	MB, #coefficient_10
	djnz	LC[0], no_rollover_20
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_20:
	move	MA, @DP[0]++
	move	MB, #coefficient_9
	djnz	LC[0], no_rollover_21
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_21:
	move	MA, @DP[0]++
	move	MB, #coefficient_8
	djnz	LC[0], no_rollover_22
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_22:
	move	MA, @DP[0]++
	move	MB, #coefficient_7
	djnz	LC[0], no_rollover_23
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_23:
	move	MA, @DP[0]++
	move	MB, #coefficient_6
	djnz	LC[0], no_rollover_24
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_24:
	move	MA, @DP[0]++
	move	MB, #coefficient_5
	djnz	LC[0], no_rollover_25
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_25:
	move	MA, @DP[0]++
	move	MB, #coefficient_4
	djnz	LC[0], no_rollover_26
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_26:
	move	MA, @DP[0]++
	move	MB, #coefficient_3
	djnz	LC[0], no_rollover_27
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_27:
	move	MA, @DP[0]++
	move	MB, #coefficient_2
	djnz	LC[0], no_rollover_28
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_28:
	move	MA, @DP[0]++
	move	MB, #coefficient_1
	djnz	LC[0], no_rollover_29
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_29:
	move	MA, @DP[0]++
	move	MB, #coefficient_0
	djnz	LC[0], no_rollover_30
	move	ACC, DP[0]
	and		#0E0h
	move	DP[0], acc
no_rollover_30:
	ret
	
;---------------------------------------------------------------------------------------------
;
;  CHECK_ADC_CHANNELS
;
; Check if ADC channels have a sample ready. If so, store the samples in the buffer.
;
check_adc_channels:
	move	C, ABF0				;See if a sample is ready
	ret		nc
;
; If sample ready bit is not set, skip this section.
;
; A sample is ready on the ADC channel. Just collect it into the circular buffer and point
; to the next sample.
;
	move	@DP[0], AD0			;Get channel 0 sample into buffer
	move	ACC, DP[0]			;Move pointer to acc
	xor		#20					;Point to other buffer
	move	DP[0], ACC
	move	@DP[0], AD1			;Get channel 1 sample into buffer
	xor		#20					;Point back to channel 0 buffer
	add		#1					;Increment buffer pointer
	cmp		#20h				;Are we beyond end?
	sjump	NE, adjust_not_required	;If not, don't adjust pointer
	move	ACC, #0				;Reset pointer back to zero
adjust_not_required:
	move	DP[0], ACC
	ret

;---------------------------------------------------------------------------------------------
;
;
; TIMER INTERRUPT SERVICE ROUTINE
;
; Implementation Notes:
;
; The MAXQ3120 operates on a 8 MHz (125 ns) clock. DS1 signals have a clock rate of 1.544 MHz
; clock (647.67 ns). That means I have slightly more than five CPU cycles between bits.
;
; That precludes testing for a clock edge. When the frame sync pulse comes, this machine has
; to be ready!
;
; It is.
;
; Note that I use two nested loops here, when one would surely be sufficient. The reason is
; that after sixteen bits, if I had only five cycles in the loop, the bit position would be
; off by:
;
; (647 ns / bit) * (16 bits)                              = 10.363 us
; (125 ns / CPU cycle) * (5 cycles per loop) * (16 loops) = 10.000 us
;                                                          ----------
;                                                              363 ns
;
; or about 1/2 bit time.
;
; The solution is to split the loop into two. That adds two cycles (250 ns) at the midpoint,
; making the system accurate to within 113 ns, or about 1/5 bit time.
;
; Of course, an external shift register would make all these timing issues go away, but the
; idea here is to use as little external hardware as possible!
;
; Note that the ISR assumes the following:
;
; 1. There are sufficient stack locations to hold two more values;
; 2. AP=APC=0 on entry;
; 3. The ISR owns LC1.
;
timer_isr:
	push	PSF					;Save flags
	push	LC[0]				;Save LC0
	move	TR0, #0				;Turn off timer interupt source
	move	LC[1], #2			;Two bytes to transfer
;
; Spin here until frame bit goes HIGH. For active low frame pulse, invert the sense of this test.
;
	move	AP, #6				;Point to A6 (character to be sent)
	move	APC, #1				;Toggle between A6 (TX char) and A7 (RX char)
wait_for_frame_pulse:
	move	C, PI2.2			;Spin here until port 2.2...
	sjump	NC, wait_for_frame_pulse	;...goes active.
;
; Frame pulse is active. Assert first data bit on TxO
;
pcm_bus_outer_loop:
	move	LC[0], #8			;Eight bits per tx/rx character
pcm_bus_main_loop:
	move	PO2, A[6]			;Output A6 (bit 0 is the only output bit on the port)
	rr							;Rotate A6 / point to A7
;
; Get data bit from RxI
;
	move	C, PI2.1			;Get input bit into carry
	rrc							;Shift C into A7 / point to A6
;
; Loop back if fewer than 16 bits
;
	djnz	LC[0], pcm_bus_main_loop
	djnz	LC[1], pcm_bus_outer_loop
;
; Clean up and return.
;
	move	T0L, #900 & 0FFh
	move	T0H, #900 << 8		;Load 900 into timer
	move	SC.3, #1			;Set done flag
	move	APC, #80h			;Point ACC back to zero / stop swapping ACC
	pop		LC[0]				;Restore LC0
	pop		PSF					;Restore flags
	reti

;---------------------------------------------------------------------------------------------
;
;
; MU-LAW TABLE
;
MU_LAW:
	dw	00000h
	dw	00005h
	dw	0000Bh
	dw	00011h
	dw	00018h
	dw	0001Fh
	dw	00026h
	dw	0002Dh
	dw	00035h
	dw	0003Dh
	dw	00045h
	dw	0004Eh
	dw	00057h
	dw	00061h
	dw	0006Bh
	dw	00075h
	dw	00080h
	dw	0008Bh
	dw	00097h
	dw	000A4h
	dw	000B1h
	dw	000BEh
	dw	000CCh
	dw	000DBh
	dw	000EAh
	dw	000FBh
	dw	0010Bh
	dw	0011Dh
	dw	0012Fh
	dw	00142h
	dw	00156h
	dw	0016Bh
	dw	00181h
	dw	00198h
	dw	001B0h
	dw	001C8h
	dw	001E2h
	dw	001FDh
	dw	0021Ah
	dw	00237h
	dw	00256h
	dw	00276h
	dw	00298h
	dw	002BBh
	dw	002DFh
	dw	00306h
	dw	0032Eh
	dw	00357h
	dw	00383h
	dw	003B1h
	dw	003E0h
	dw	00412h
	dw	00446h
	dw	0047Ch
	dw	004B4h
	dw	004EFh
	dw	0052Dh
	dw	0056Dh
	dw	005B0h
	dw	005F7h
	dw	00640h
	dw	0068Ch
	dw	006DCh
	dw	00730h
	dw	00787h
	dw	007E2h
	dw	00841h
	dw	008A4h
	dw	0090Ch
	dw	00978h
	dw	009E9h
	dw	00A5Fh
	dw	00ADBh
	dw	00B5Bh
	dw	00BE2h
	dw	00C6Eh
	dw	00D01h
	dw	00D9Ah
	dw	00E3Ah
	dw	00EE1h
	dw	00F8Fh
	dw	01045h
	dw	01103h
	dw	011CAh
	dw	01299h
	dw	01372h
	dw	01454h
	dw	01540h
	dw	01636h
	dw	01738h
	dw	01845h
	dw	0195Dh
	dw	01A83h
	dw	01BB5h
	dw	01CF5h
	dw	01E42h
	dw	01F9Fh
	dw	0210Bh
	dw	02287h
	dw	02415h
	dw	025B3h
	dw	02764h
	dw	02928h
	dw	02B01h
	dw	02CEEh
	dw	02EF1h
	dw	0310Ah
	dw	0333Ch
	dw	03586h
	dw	037EBh
	dw	03A6Ah
	dw	03D06h
	dw	03FBFh
	dw	04297h
	dw	04590h
	dw	048AAh
	dw	04BE7h
	dw	04F49h
	dw	052D2h
	dw	05682h
	dw	05A5Ch
	dw	05E62h
	dw	06296h
	dw	066F9h
	dw	06B8Eh
	dw	07056h
	dw	07555h
	dw	07A8Dh

;---------------------------------------------------------------------------------------------
;
;
; A-LAW TABLE
;
A_LAW:
	dw	00000h
	dw	0000Fh
	dw	0001Fh
	dw	0002Fh
	dw	0003Fh
	dw	0004Fh
	dw	0005Fh
	dw	0006Fh
	dw	0007Fh
	dw	0008Fh
	dw	0009Fh
	dw	000AFh
	dw	000BFh
	dw	000CFh
	dw	000DFh
	dw	000EFh
	dw	000FFh
	dw	0010Fh
	dw	0011Fh
	dw	0012Fh
	dw	0013Fh
	dw	0014Fh
	dw	0015Fh
	dw	0016Fh
	dw	0017Fh
	dw	00190h
	dw	001A2h
	dw	001B4h
	dw	001C7h
	dw	001DBh
	dw	001F0h
	dw	00205h
	dw	0021Ch
	dw	00234h
	dw	0024Ch
	dw	00266h
	dw	00281h
	dw	0029Dh
	dw	002BAh
	dw	002D9h
	dw	002F9h
	dw	0031Ah
	dw	0033Ch
	dw	00361h
	dw	00386h
	dw	003AEh
	dw	003D7h
	dw	00402h
	dw	0042Fh
	dw	0045Eh
	dw	0048Fh
	dw	004C2h
	dw	004F7h
	dw	0052Eh
	dw	00568h
	dw	005A5h
	dw	005E4h
	dw	00626h
	dw	0066Bh
	dw	006B2h
	dw	006FDh
	dw	0074Bh
	dw	0079Dh
	dw	007F2h
	dw	0084Bh
	dw	008A8h
	dw	00909h
	dw	0096Eh
	dw	009D7h
	dw	00A45h
	dw	00AB8h
	dw	00B30h
	dw	00BADh
	dw	00C30h
	dw	00CB8h
	dw	00D46h
	dw	00DDBh
	dw	00E76h
	dw	00F17h
	dw	00FC0h
	dw	01070h
	dw	01128h
	dw	011E8h
	dw	012B0h
	dw	01381h
	dw	0145Bh
	dw	0153Fh
	dw	0162Dh
	dw	01725h
	dw	01827h
	dw	01936h
	dw	01A50h
	dw	01B76h
	dw	01CA9h
	dw	01DE9h
	dw	01F38h
	dw	02095h
	dw	02202h
	dw	0237Eh
	dw	0250Bh
	dw	026A9h
	dw	02859h
	dw	02A1Dh
	dw	02BF4h
	dw	02DDFh
	dw	02FE0h
	dw	031F7h
	dw	03426h
	dw	0366Dh
	dw	038CEh
	dw	03B49h
	dw	03DE0h
	dw	04094h
	dw	04367h
	dw	04658h
	dw	0496Bh
	dw	04CA0h
	dw	04FF9h
	dw	05377h
	dw	0571Dh
	dw	05AEBh
	dw	05EE4h
	dw	06309h
	dw	0675Ch
	dw	06BE0h
	dw	07097h
	dw	07582h
	dw	07AA4h

end
